//
// OdfObject.cc
//
// Giuseppe Vacanti, European Space Agency and cosine science & computing
// (c) 1998-2001
//
//  The main method is scan(). This finds all correctly named ODF
//  components in directory. Then each component is examined and its
//  main parameters (instrument, date start, date end) extracted and
//  used to build the ODF summary file.
//
//  file_set -- set tallying the names of the dataset that have
//  already been processed. Can be used to prevent processing of some
//  datasets. 
//
//  file_list -- these datasets will be listed in the ODF summary
//  file. Can be used to add files that are not DAL datasets.
//

#include <string>
#include <vector>
#include <map>
#include <set>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <algorithm>

#include <cstdlib>
#include <cassert>

#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <Dal.h>
#include <param.h>
#include <Xmm.h>
#include <CalServer.h>
#include <Odf.h>
#include <STime.h>
#include <stringutil.h>
#include <SasDalUtils.h>
#include <Directory.h>

#include<TimeOrderedTable.h>

#include "ExposureObject.h"
#include "OdfObject.h"
#include "OdfConstituent.h"

#include "IppvExtractor.h"

// Why are these not members?
static unsigned global_expo_counter = 0;
static map<string,unsigned> instrument_expo_counter;

// My full name via the SAS infrastructure.
static string me() {
	PROC RETURN (string(Info::Task::current()->name()) + "(" + string(Info::Task::current()->package()) + "-" + string(Info::Task::current()->version()) + ") [" + string(Info::Task::current()->release()) + "]");
}


static inline string expid(const string& inst, const string& exp){
	return inst + "/" + exp;
}

OdfObject::OdfObject(const string& dir, 
	const bool findinstrumentmodes,
	const bool usehousekeeping)	: 
		start_time(STime::latest),
		end_time(STime::earliest),
		_dir(dir),
		_find_instrument_modes(findinstrumentmodes),
		_use_housekeeping(usehousekeeping),
		_valid(false),
		_tpad(240.)
		// 2000-11-13: I have observed up to 90s delay between DATE-OBS
		// and when an HK parameter actually changes value. 
		// 2002-01-28: there is at least one HK parameter (MOS) that is 
		// updated every 208s. Use 240s because it is 4 minutes.
{
	PROC

	// There can be three diffrent ways to call one instrument
	instruments.push_back("M1");
	long_names["M1"] = "EPIC MOS";
	fits_names["M1"] = "EMOS1";

	instruments.push_back("M2");
	long_names["M2"] = "EPIC MOS";
	fits_names["M2"] = "EMOS2";

	instruments.push_back("R1");
	long_names["R1"] = "RGS";
	fits_names["R1"] = "RGS1";

	instruments.push_back("R2");
	long_names["R2"] = "RGS";
	fits_names["R2"] = "RGS2";

	instruments.push_back("PN");
	long_names["PN"] = "EPIC PN";
	fits_names["PN"] = "EPN";

	instruments.push_back("OM");
	long_names["OM"] = "OM"; 
	fits_names["OM"] = "OM";

	// Data modes that contain science data. These are used to determine
	// the camera mode.
	science_data_modes.push_back(Xmm::IMAGING);
	science_data_modes.push_back(Xmm::TIMING);
	science_data_modes.push_back(Xmm::REDUCED_IMAGING);
	science_data_modes.push_back(Xmm::COMPRESSED_TIMING);
	science_data_modes.push_back(Xmm::BURST);
	science_data_modes.push_back(Xmm::DATA_SPECTROSCOPY);
	science_data_modes.push_back(Xmm::HIGH_TIME_RESOLUTION);
	science_data_modes.push_back(Xmm::DATA_FAST);

	// Data modes for which HK parameter extraction is not
	// attempted. This list must not contain any science data mode, but
	// only "auxiliary" data modes for which HK extraction is known to
	// trigger inappropriate warnings. Write why a data mode is being
	// excluded.

	// DISCARDED_LINES_DATA
	// These data files are often sent to the ground after the end of
	// the observation. For this reason they are known to contain the
	// wrong DATE-END. When DATE-END is past the last HK record
	// available, many warnings are issued that only confuse the user. 
	_no_ippv_data_modes.push_back(Xmm::DISCARDED_LINES_DATA);
	_no_ippv_data_modes.push_back(Xmm::HBR_OFFSET_DATA);


	scan();

	if(file_list.size() > 0)
		_valid = true;

	ENDPROC
}  

string OdfObject::identifier() const {
	PROC
	RETURN revolution + "_" + proposal + observation_id;
}

// This is the name used by the SAS
string OdfObject::canonicalSummaryName() const {
	PROC
	RETURN revolution + "_" + proposal + "_SCX00000SUM.SAS";
}

// This is the name of the ODS-generated summary file.
string OdfObject::icdosf() const {
	PROC
	RETURN revolution + "_" + proposal + "_SCX00000SUM.ASC";
}

// Use these two methods to manipulate ascii ODF components.
void OdfObject::insertFile(const string name) {
	PROC
	file_list.push_back(name);
	ENDPROC
}

void OdfObject::removeFile(const string name) {
	PROC
	file_list.erase(find(file_list.begin(), file_list.end(), name));
	ENDPROC
}


// Identify all ODF constituents.
void OdfObject::scan() {
	PROC

	bool odfNotKnown = true; 	// This indicates that we do not yet know
								// the ODF identifier.

	list<STime> stimes;

	Directory d(_dir);
	typedef Directory::const_iterator LI;

	// Determine from hk the maximum allowed DATE-END for each
	// instrument. Aim to use:
	//	PEH file for MOS and OM
	//	PFH file for RGS
	// 	PMH file for PN
	//
	d.filter("????_??????????_??X00000PEH.FIT|????_??????????_??X00000PFH.FIT|????_??????????_??X00000PMH.FIT");
	std::map<Xmm::Instrument, STime> last_times;

	for(LI j = d.begin(); j != d.end(); ++j) 
	{
		const DirEntry& e = **j;
		OdfSet odfset(e.basename());
		DataSet * set = dataSetServer->open(e.name(), Dal::Read,
						DataSetServer::Low);
		TimeOrderedTable * tot = new TimeOrderedTable(set->table(e.name() + ":"));
		last_times[odfset.instrument()] = tot->last_time();
		delete tot;
		dataSetServer->close(set);
	}

	errstr << "The following are the latest possible exposure end dates by instrument as derived from the main housekeeping datasets: " << std::endl;

	for(std::map<Xmm::Instrument, STime>::iterator j = last_times.begin();
		j != last_times.end(); 
		++j) 
	{
		errstr << (*j).first << " " 
			<< (*j).second.toString(0) 
			<< std::endl;
	}
	
	errstr << Sas::message(Msg::AppLib, Msg::Verbose);

	// latest DATE-END from HK [end] -------------------------------------------

	// First look only at the *.FIT files flagged as being Scheduled or
	// Unscheduled. These define the observation structure in terms of
	// exposures and active istruments.

	d.filter("????_??????????_??S????????.FIT|????_??????????_??U????????.FIT");
	
	errstr << "Looking for ODF constituents in " + d.name()
		<<  Sas::message(Msg::App,Msg::Sparse);

	multimap<string,ExposureObject,less<string> > hdi_exp_time_map;
		
	for(LI j = d.begin(); j != d.end(); j++) 
	{

		bool useTime = true;

		const DirEntry& e = **j;

		// *.FIT files are DAL datasets. Put them in the list.

		pair<set<string, less<string> >::const_iterator, bool> p =
	  		file_set.insert(e.name());
			
		if (p.second) 
		{ // This is a file we have not yet seen
	  		errstr << "Examining ODF constituent " + e.basename() 
				<< Sas::message(Msg::App,Msg::Noisy);

			// FIXME: OdfConstituent should be implemented in terms of OdfSet.
			OdfConstituent oc(e.name());
			OdfSet odfset(e.basename());

			// FIXME: this bit of code is duplicated below
			if(odfNotKnown)
			{
				odfNotKnown = false;

				// Initialize the observation ID, the revolution number, and
				// the proposal ID. These are unique to one ODF, so we take the
				// first one we find.

				observation_id = oc.observation();
				revolution = oc.revolution();
				proposal = oc.proposal();

				errstr << "Identified the following ODF" << endl;
				errstr << "Revolution:  " << revolution << endl;
				errstr << "Observation: " << observation_id << endl;
				errstr << "Proposal:    " << proposal
					   << Sas::message(Msg::App, Msg::Silent);
			} 
			else 
			{

				// From the second constituent on check that it is the very
				// same observation. 

				if(observation_id != oc.observation() ||
					revolution != oc.revolution() ||
					proposal != oc.proposal()) 
				{
					errstr << "The following observation was identified: " +
					identifier() + ". But this dataset " + oc.base_name() +
					" belongs to a different observation." <<
					Sas::error("MultipleObservations");
				}
			}

			typedef pair<const string, OdfConstituent>  SOC;
			typedef pair<const string, string> SS;

			file_list.push_back(oc.base_name());      
			SOC pf(oc.instrument(),oc);
			file_map_by_inst.insert(pf);      
			SS pm(oc.instrument(),oc.long_exposure());
			exposure_map_by_inst.insert(pm);

			DataSet * set = dataSetServer->open(oc.name(), DataSet::Read,
							DataSetServer::Low);

			// Access by number as some ODF components may not have block
			// names. Also block names encode the block number ... sigh!

			ExposureObject this_exp;
			this_exp.exposure = oc.exposure();
			this_exp.scheduled = oc.scheduled();

			// For each dataset we want to determine a  time start/stop,
			// from which the start/stop of the exposure will be
			// derived. There are several problems we can encounter, and we
			// deal with them below.

			Block* blk = set->block(0); 

			// DATE-OBS gotta be right. A wrongly formatted string will make
			// us crash.
			this_exp.start = STime(blk->attribute("DATE-OBS")->asString());


			// FIXME: EPN discarded lines data sets have the wrong DATE-END
			// This arises because often the data are set to the ground
			// after the observation has been completed. This can also be
			// hours! DATE-START is correct.
			// If the wrong DATE-END was used, the observation would
			// appear to be much longer than it actually is. Moreover, a
			// task that is asking the OAL how long the observation is and
			// then used this info to make use of the attitude data, would
			// discover that for most of the observation there is no valid
			// attitude. Warnings ensue...
			// Solution: Impose a 2*_tpad duration for each exposure.
			// FIXME: same treatment for RGS DII


			// 2002-02-14: decide instead to use 1s as the arbitrary
			// duretion. Rationale: for exposures containing both science
			// and diagnostic data, adding 2*tpad (480s) to DATE-END of some
			// of the components may alter the exposure duration calculation
			// for the HK extraction software. This in turn can lead to some
			// HK parameters not being extracted correctly (because we end
			// up looking too much after the end of the exposure).
			if(	odfset.dataMode() == Xmm::DISCARDED_LINES_DATA ||
				odfset.dataMode() == Xmm::DATA_DIAGNOSTIC)
			{
				errstr << "Detected a data set with data mode "
				   << odfset.dataMode() << ". This is know to contain the wrong DATE-END. I'm assigning to it an arbitrary duration of 1 second."
		   			<< Sas::message(Msg::App, Msg::Silent);
				this_exp.stop = this_exp.start + 1.;
			} 
			// HBR_OFFSET_DATA is known to occur outside the
			// observation window. Do not include it in the time list
			else if(odfset.dataMode() == Xmm::HBR_OFFSET_DATA) 
			{

					errstr << "Detected a data set with data mode "
						<< odfset.dataMode() << ". These are dumped outside the observation windows and so have wrong DATE-OBS and DATE-END values. I'll set the times to match the last time reported by the housekeeping."
						<< Sas::message(Msg::App, Msg::Silent);

					this_exp.start = last_times[odfset.instrument()] - 1;
					this_exp.stop  = last_times[odfset.instrument()];

					useTime = false;

			}
			else 
			{
				// We know that DATE-END can be wrong in one of more
				// manner. We have seen (at least):
				//   . empty
				//   . -1
				//   . 0000-00-00T00:00:00
				// Turn on exceptions in STime, so we get a chance to recover.

				STime::errorHandling = STime::EXCEPTION; 
				try 
				{
					this_exp.stop = STime(blk->attribute("DATE-END")->asString());
				}		
				catch(STime::InvalidTime ex)
				{
					errstr
					<< "An error was detected while trying to determine DATE-END for "
					<< e.basename() << ". The error said: " << ex.arg() << ".";

					// First we try to use the maximum allowed DATE-END from the
					// HK.

					if( last_times.find(odfset.instrument()) !=
						last_times.end())
					{
						this_exp.stop = last_times[odfset.instrument()];
						errstr << " I'll use a DATE-END = " << this_exp.stop.toString(0)
							<< " as determined from the main HK dataset.";
					} 
					else 
					{
						// Use an arbitrary (short) duration.
						this_exp.stop = this_exp.start + 2*_tpad;	    
						errstr << "This dataset will be assigned an arbitrary duration of "
							<< 2*_tpad << " seconds.";
	  				}

					errstr << Sas::message(Msg::AppLib, Msg::Silent);
				}

				STime::errorHandling = STime::ERRORANDEXIT;

				// We have seen several cases when DATE-END is a few days
				// after DATE-OBS. This is caused by quirks in the ground
				// segment software, but will confuse us when we try to
				// extract HK values for this exposure. To prevent the latter,
				// we reset the end time to match the value found in the HK
				// datset.
				if(last_times.find(odfset.instrument()) != last_times.end())
				{
				
					if(this_exp.stop > last_times[odfset.instrument()])
					{

						errstr << "DATE-END = " << this_exp.stop.toString(0)
							<< " as reported by " << e.basename()
							<< " is later than the last available housekeeping record "
							<<  last_times[odfset.instrument()].toString(0)
							<< " I'll set DATE-END to match the date reported by the housekeeping."
							<< Sas::message(Msg::AppLib, Msg::Silent);
						this_exp.stop = last_times[odfset.instrument()];

						// SPR 2592 Make sure that DATE-OBS is before DATE-END
						if(this_exp.start >= this_exp.stop)
						{

							errstr << "DATE-OBS = " << this_exp.start.toString(0)
								<< " as reported by " << e.basename()
								<< " is later than the new DATE-END record "
								<<  this_exp.stop.toString(0)
								<< " I'll set DATE-OBS to before DATE-END."
								<< Sas::message(Msg::AppLib, Msg::Silent);
							this_exp.start = this_exp.stop - 1;
						}

					}
					
				} 
				else 
				{
					errstr << "I don't seem to have been able to determine the time of the last housekeeping record for "
						<< odfset.instrument()
						<< Sas::message(Msg::AppLib, Msg::Silent);
				}

				if(this_exp.start == this_exp.stop)
				{

					if (odfset.dataMode() == Xmm::AUXILIARY)
					{
						errstr << "Detected a " << odfset.instrument() 
								<< " dataset with data mode "
								<< odfset.dataMode() 
								<< " with zero duration. This indicates that this exposure is unlikely to processable by the SAS."
								<< Sas::warning("ZeroExposureLength");
					}

				}

			}

			errstr << "Dataset Time Range: " << oc.base_name() << " : " << this_exp.start << " - "
				<< this_exp.stop << Sas::message(Msg::App, Msg::Sparse);

			stimes.push_back(this_exp.start);
			stimes.push_back(this_exp.stop);
			
			// SCR 117 - need to process HDI files after the rest			
			pair<const string, ExposureObject> pe(oc.instrument(), this_exp);
			if (useTime) {	
				inst_exp_time_map.insert(pe);
			}
			else {
				hdi_exp_time_map.insert(pe);
			}
			
			dataSetServer->close(set);

		} // I have this file already?
		
	} // loop over all directory entries
	
	
	// Update time_exp map with solitary HDI exposures
	multimap<string,ExposureObject>::const_iterator hdi_iter = 
		hdi_exp_time_map.begin();

	for (; hdi_iter != hdi_exp_time_map.end(); hdi_iter++) {

		bool updateTimes = true;

		ExposureObject hdiexp = (*hdi_iter).second;

		multimap<string,ExposureObject>::const_iterator exp_iter = 
			inst_exp_time_map.find((*hdi_iter).first);

		for (; exp_iter != inst_exp_time_map.end() && updateTimes; exp_iter++) {
			ExposureObject instexp = (*exp_iter).second;
		
			if ((hdiexp.exposure  == instexp.exposure) &&
				(hdiexp.scheduled == instexp.scheduled)) {
				updateTimes = false;						
			}
			
		}
		
		if (updateTimes) {
			errstr << "Detected a solitary HBROffset data set. Providing correct exposure information"
				<< Sas::message(Msg::App, Msg::Silent);
			pair<const string, ExposureObject> pe((*hdi_iter).first, hdiexp);
			inst_exp_time_map.insert(pe);				
		}
		else {
			errstr << "Detected an HBROffset data set within a scientific exposure which not need separate exposure information"
				<< Sas::message(Msg::App, Msg::Silent);
		}
		
	}

	// Count up also the ascii files and the FITS files with X (neither
	// scheduled nor unscheduled) flag. These FITS files are not used to
	// determine the observation duration.
	d.filter("????_??????????_???????????.ASC|????_??????????_??X????????.FIT");

	for(LI j = d.begin(); j != d.end(); j++) 
	{
		const DirEntry &e = **j;
		OdfConstituent oc(e.name());
		
		if(odfNotKnown)
		{
			odfNotKnown = false;
			observation_id = oc.observation();
			revolution = oc.revolution();
			proposal = oc.proposal();

			errstr << "Identified the following ODF" << endl;
			errstr << "Revolution:  " << revolution << endl;
			errstr << "Observation: " << observation_id << endl;
			errstr << "Proposal:    " << proposal
				<< Sas::message(Msg::App, Msg::Silent);
		} 
		else 
		{
			if(observation_id != oc.observation() ||
				revolution != oc.revolution() ||
				proposal != oc.proposal()) 
			{
				
				errstr << "The following observation was identified: " +
					identifier() + ". But this dataset " + oc.base_name() +
					" belongs to a different observation." <<
					Sas::error("MultipleObservations");
			}
		}
		
		file_list.push_back(e.basename());      
		
	}

	if(stimes.size() == 0 || odfNotKnown) 
	{
		_valid = false;
	}
	else 
	{
		stimes.sort();
		start_time = *stimes.begin();
		end_time = *(--stimes.end());
		errstr << "Observation Time Range: " << start_time << " - " << end_time << Sas::message(Msg::App, Msg::Silent);
	}

	ENDPROC
}

void OdfObject::writeSummary(ostream & os,
			     const bool writepath) const {
	PROC

	if(_valid)
	{

		os << "// ----------------------------------" << endl;	
		os << "// XMM-Newton Science Analysis System" << endl;
		os << "// ----------------------------------" << endl;
		os << "//" << endl;
		os << "// ODF Summary File" << endl;
		os << "// By: " << me() << " on " << STime("now") << endl;
		os << "//" << endl;
		
		if(writepath)
		{
			os << "//" << endl;
			os << "// Directory where the ODF constituents were found. This may have to be edited to match the local file system structure." << endl;
			os << "//" << endl;
			os << "PATH " << _dir << endl;
		}

		observation_rec(os);
		file_details_rec(os);
		configuration_details_rec(os);

		typedef vector<string>::const_iterator VI;
		for(VI i = instruments.begin(); i != instruments.end(); i++) 
		{
			string inst = *i;
			instrument_details_rec(inst, os);
			typedef multimap<string,string,less<string> >::const_iterator MI;
			MI lb = exposure_map_by_inst.lower_bound(inst);
			MI ub = exposure_map_by_inst.upper_bound(inst);
			set<string, less<string> > exposures;
			for(MI j = lb; j != ub; j++) 
			{
				pair<set<string, less<string> >::const_iterator, bool> p = 
				exposures.insert((*j).second);
				if(p.second) 
				{
					exposure_sub_record(inst, (*j).second, os);
					configuration_sub_record(inst, (*j).second, os);
				}
			}
		}

		proposal_summary_rec(os);

	} 
	else 
	{
		errstr << "Could not find a valid ODF in " << _dir
			<< ", hence the summary file will not be written."
			<< Sas::error("NoOdfFound");
	}

	ENDPROC

}

string OdfObject::hkfile(const string& i, const string& type)
{
	// Code the ODF ICD naming convention
	PROC
	RETURN revolution + "_" + proposal + "_" + i +"X" + "00000" + type + "H.FIT";
}  

void OdfObject::observation_rec(ostream & os) const {
	PROC
	errstr << "Writing the observation record." <<
		Sas::message(Msg::App, Msg::Noisy);

	// We are going to update the observation start/end with possibly
	// more accurate data as they are extracted from the HK and/or
	// adjusted in various manners. All the work is done in
	// exposures_start_end_times.

	vector<STime> t;
	typedef multimap<string,OdfConstituent,less<string> >::const_iterator MI;
	for(unsigned k=0; k<instruments.size(); ++k)
	{
		MI lb = file_map_by_inst.lower_bound(instruments[k]);
		MI ub = file_map_by_inst.upper_bound(instruments[k]);  
		for(MI i = lb; i != ub; i++) 
		{
			vector<STime> v = exposure_start_end_times(instruments[k],
				(*i).second.long_exposure());
			t.push_back(v[0]);
			t.push_back(v[1]);
		}
	}

	// Normally we never enter the following branch, as there is at least one
	// science exposure per ODF. However we may be processing SDFs, where there
	// can be only HK files. In this case we try to determine the observation 
	// duration by looking at whatever datasets are available in
	// file_list. Normally these would be the HK files.

	if(t.size() == 0)
	{
	    Sas::warning("NoScienceFiles", "There appear to be no science files in this observation. I'll try to determine its duration from whatever other files I could find.");
		if(file_list.size() == 0) 
		{
			Sas::error("NoFiles", "There appear to be no files of any kind associated with this observation.");
		}
		
		for(unsigned k=0; k<file_list.size(); ++k)
		{
			string f(_dir + "/" + file_list[k]);
			if(dataSetServer->exists(f))
			{
				DataSet* set = dataSetServer->open(f, DataSet::Read,
								DataSetServer::Low);
				Block* blk = set->block(0);
				STime::errorHandling = STime::EXCEPTION;
				// Here we require that both DATE-OBS and DATE-END be valid.
				try 
				{
					STime t1 = STime(blk->attribute("DATE-OBS")->asString());
					STime t2 = STime(blk->attribute("DATE-END")->asString());
					errstr << f << " start/stop: " << t1 << "/" << t2
						 << Sas::message(Msg::AppLib, Msg::Sparse);
					t.push_back(t1);
					t.push_back(t2);
				}
				catch(...)
				{	}
				STime::errorHandling = STime::ERRORANDEXIT;
				dataSetServer->close(set);
			}
		}
		if(t.size() == 0)
		{
			errstr << "There appear to be no files associated with this ODF from which a valid start/stop interval could be extracted. This is the list of files I have:" << endl;
			for(unsigned k=0; k<file_list.size(); ++k) {
				errstr <<  file_list[k] << endl;
			}
			errstr << Sas::error("NoFiles");
		}
	}

	sort(t.begin(), t.end());
	start_time = t[0];
	end_time = t[t.size()-1];

	os << "//" << endl;
	os << "// Observation Record" << endl;
	os << "//" << endl;

	// FIXME: SLEWS have 9RRRR#####
	//        unscheduled observations are 99999#####
	// I (and the ground segment) rely on the fact that we are unlikely
	// to reach 9999 revolutions :) A more accurate test would be
	// observation_id.substring(0,5) == "9RRRR"
	os << ((observation_id[0] == '9' && observation_id[1] != '9') ? "SLEW" : "OBSERVATION") << endl;
	os << observation_id << " / Observation Identifier" << endl;
	os << revolution << " / Revolution Identifier" << endl;
	os << start_time << " / Observation Start Time" << endl;
	os << end_time << " / Observation End Time" << endl;

  ENDPROC
}

void OdfObject::file_details_rec(ostream & os) const {
	PROC
	errstr << "Writing the file details record." <<
		Sas::message(Msg::App, Msg::Noisy);

	// FIXME: sorting file_list does not compile??!
	vector<string> v(file_list);
	sort(v.begin(), v.end());
	//  sort(file_list.begin(), file_list.end());
	os << "//" << endl;
	os << "// Number of files in the Observation Summary File, and their names" << endl;
	os << "//" << endl;
	os << "FILES" << endl;
	os << file_list.size() << " / Number of Files" << endl;
	
	typedef vector<string>::const_iterator VI;
	//  for(VI i = file_list.begin(); i != file_list.end(); i++) {
	for(VI i = v.begin(); i != v.end(); i++) 
	{
		const string &s = *i;
		os << s << " / ODF Constituent" << endl;
	}
	
	ENDPROC
}

void OdfObject::configuration_details_rec(ostream& os) const {
	PROC
	errstr << "Writing the configuration details record." <<
		Sas::message(Msg::App, Msg::Noisy);

	// FIXME: code duplication. Also in proposal_rec()
	Directory d(_dir);
	d.filter(icdosf());
	if(d.size() != 1) 
	{
		errstr << "Looking for one and only one file named " << icdosf()
			<< " in the directory " << _dir << ". Instead I found "
			<< d.size() << Sas::error("IncorrectNumberOfDatasets");
	}

	// FIXME: stream status never checked (also in proposal_rec)
	// On the other hand, this file has been read by the OAL already.
	ifstream is((*d.begin())->name().c_str());
	string line;
	while(getline(is,line) && line.substr(0,13) != "CONFIGURATION" && is.good());

	if(!is.eof()){

		os << "// The CONFIGURATION record was extracted from " << (*d.begin())->basename()
			<< endl;
			
		os << line << endl;

		// Now I expect to find a number
		getline(is,line);
		unsigned nconf = 0;

		try 
		{
			nconf = toNumber<unsigned>(line);
		}
    	catch(string & e) 
		{
			Sas::message(Msg::App, Msg::Silent, "Trying to convert '" + line
				+ "' to a number failed with the message'" +
				e + "' while reading the CONFIGURATION record. Will assume that there are no configuration records.");
			nconf = 0;
    	}
		os << nconf + 3 << endl; // Add additional odfingest metadata records

		for(unsigned k = 0; k < nconf  && is.good(); ++k)
		{
			getline(is,line);
			os << line << endl;;
		}
		
		// Add odfingest metadata
		os << "// odfingest metadata" << endl;
		os << "ODFINGEST_VERSION = " << string(Info::Task::current()->version()) << " / odfingest version" << endl;
		os << "SAS_MANIFEST = " << string(Info::Task::current()->release()) << " / SAS Manifest" << endl;
		os << "PROCESS_DATE = " << STime("now") << " / Time and Date when summary file produced" << endl;

	} 
	else {
		Sas::message(Msg::App, Msg::Silent,
		"No CONFIGURATION record found. This indicates that the ODF was created before ~ April 2002, and does not stop the SAS from processing your data.");
	}	

	is.close();
	ENDPROC

}

bool OdfObject::is_instrument_active(const string& inst) const {
  PROC
  bool a = false;
  if(number_of_exposures_of_instrument(inst) > 0) a = true;
  RETURN a;
}

unsigned int OdfObject::number_of_exposures_of_instrument(const string& inst) const {
	PROC

	// Here equal_range should work, but I cannot make it work.
	typedef multimap<string,OdfConstituent,less<string> >::const_iterator MI;
	MI lb = file_map_by_inst.lower_bound(inst);
	MI ub = file_map_by_inst.upper_bound(inst);  
	set<string,less<string> > s;  

	for(MI i = lb; i != ub; i++) {
		s.insert((*i).second.long_exposure());
	}

	RETURN s.size();
}

void OdfObject::instrument_details_rec(const string& inst,
				       ostream & os) const {
	PROC
	unsigned int n = number_of_exposures_of_instrument(inst);
	string active = "Y";
	if(n == 0) active = "N";    
	os << "//" << endl;
	os << "// Instrument Record" << endl;
	os << "//" << endl;
	os << "INSTRUMENT" << endl;
	os << inst << endl;
	os << active << " / Is this instrument active?" << endl;
	os << n << " / Number of exposures for this instrument" << endl;
	ENDPROC
}

const list<STime> OdfObject::exposure_times(const string& inst,
					    const string& exp) const
{
	PROC

	// FIXME: another contorsion
	// We rely on the caller using the long_exposure.
	assert(exp.size() == 4);
	assert(exp[0] == 'S' || exp[0] == 'U');

	typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
	EI lb = inst_exp_time_map.lower_bound(inst);
	EI ub = inst_exp_time_map.upper_bound(inst);
	list<STime> stimes;

	for(EI i = lb; i != ub; i++) 
	{
    	const  ExposureObject &e = (*i).second;
    	// FIXME: The exposure string is now '?nnn', where ? is one of 'S'
    	// 'U', so we have to combine two strings from the ExposureObject,
    	// pending a better design.
		if(e.scheduled + e.exposure == exp) 
		{
			stimes.push_back(e.start);
			stimes.push_back(e.stop);	
		}
	}
	
	stimes.sort();

	RETURN stimes;
}

vector<STime> OdfObject::exposure_start_end_times(const string& inst,
						  const string& exp) const
{
	PROC

	list<STime> stimes = exposure_times(inst,exp);

	STime exp_dstart = *stimes.begin();
	STime exp_dend = *(--stimes.end());

	// This is OM, and we know that DATE-OBS and DATE-END are not
	// reliable. Try to find the good values in the NPHK.
	if(inst == "OM")
	{
		Directory d(_dir);
		d.filter("????_??????????_OM??????NPH.FIT");    
		if(d.size() != 1)
		{
		errstr << "Looking for one and only one file matching the pattern "
			 << d.filter() << " (OM non periodic housekeeping) in the directory "
			 << _dir << ". I found " << d.size() << " instead. The lack of OM non periodic housekeeping may be a problem later in the SAS. Here it makes me use incorrect start/stop times for the observation."
			 << Sas::warning("IncorrectNumberOfDatasets");
		} 
		else 
		{

			DataSet *set = dataSetServer->open((*d.begin())->name(), Dal::Read);
			Table *tab = set->table(0);
			ColumnData *sid = tab->column("SID")->data();
			int16 *v = sid->int16Data();
			TimeOrderedTable *t = new TimeOrderedTable(set->table(0));
			unsigned mark=0;
			for(unsigned k=0; k<t->rows(); ++k)
			{
				// cout << k << " " << v[k] << endl;
				// Skip if the time tag is not defined: SOC-SPR-2095
				if(!t->row_has_time(k)) {
					continue;
				}
				
				if(t->time_of_row(k) > exp_dstart)
				{
					mark = k;
					// cout << "Will start looking at row " << mark << endl;
					break;
				}
			}

			bool have148 = false;
			bool have149 = false;

			for(unsigned k=mark; k<t->rows(); ++k)
			{
				//      cout << k << " " << v[k] << endl;
				// Skip if the time tag is not defined: SOC-SPR-2095
				if(!t->row_has_time(k)) {
					continue;
				}
				
				if(have148 && have149) {
					break;
				}

				if(v[k] == 148)
				{
					if (!have148) 
					{
						exp_dstart = t->time_of_row(k);
						// cout << "START " << k << " " << v[k] << t->time_of_row(k) << endl;
						have148 = true;
					}
					else
					{
						// Take this rows time as the latest time for an exposre
						// exp_dend = t->time_of_row(k);
						// OR
						// Time passed into routine likely to be correct? (do nothing)
						errstr << "OM Exposure " << expid(inst, exp) << " does not have correct EXP_START/EXP_END pairs, using "
							<< exp_dend << " for Actual End Time" 
							<< Sas::warning("noOMExposurePair");
						have149 = true;
					}
				}
				
				if(v[k] == 149)
				{
					exp_dend = t->time_of_row(k);
					// cout << "STOP " << k << " " << v[k] << t->time_of_row(k) << endl;
					have149 = true;
				}
			}

			delete sid;
			delete t;
			dataSetServer->close(set);
		}
	}

	vector<STime> v(2);
	v[0] = exp_dstart;
	v[1] = exp_dend;

	errstr << expid(inst, exp) << " start = " << exp_dstart
		<< " end = " << exp_dend << Sas::message(Msg::App, Msg::Sparse);

	RETURN v;
}

void OdfObject::exposure_sub_record(const string& inst,
				    const string& exp,
				    ostream & os) const {
	PROC
	errstr << "Writing the exposure sub-record for " << expid(inst,exp)
		<< Sas::message(Msg::App, Msg::Noisy);

	// FIXME: another contorsion
	// We rely on the caller using the long_exposure.
	assert(exp.size() == 4);
	assert(exp[0] == 'S' || exp[0] == 'U');
	bool scheduled = (exp[0] == 'S');

	vector<STime> v = exposure_start_end_times(inst,exp);
	STime exp_dstart = v[0];
	STime exp_dend = v[1];

	string proposal_exp_id = proposal + inst + exp;
	string commanded_exp_id = proposal_exp_id;

	os << "//" << endl;
	os << "// Exposure Record" << endl;
	os << "// Observation Level Exposure Index = " << ++global_expo_counter << endl;
	os << "// Instrument " << "[" << inst << "] Level Exposure Identifier = " << exp << endl;
	os << "// Instrument " << "[" << inst << "] Level Exposure Index = " << ++instrument_expo_counter[inst] << endl;
	os << "//" << endl;
	os << "EXPOSURE" << endl;
	os << exp.substr(1,4) << " / Exposure Identifier [also " << exp <<"]" << endl;
	os << (scheduled ? "Y" : "N") << " / Is this a scheduled exposure?" << endl;
	os << "SCIENCE" << " / Exposure Type" << endl;
	os << commanded_exp_id << " / Commanded Exposure Identifier" << endl;
	os << proposal_exp_id << " / Proposal Exposure Identifier" << endl;
	os << exp_dstart << " / Scheduled Start Time" <<endl;
	os << exp_dend << " / Scheduled End Time" << endl;
	os << exp_dstart << " / Actual Start Time" << endl;
	os << exp_dend << " / Actual End Time" << endl;

	ENDPROC
}

void OdfObject::configuration_sub_record(const string& inst,
					 const string& exp,
					 ostream & os) const {


	PROC
	errstr << "Writing the configuration sub-record for "
		<< expid(inst,exp)
		<< Sas::message(Msg::App, Msg::Noisy);

	// FIXME: another contorsion
	// We rely on the caller using the long_exposure.
	assert(exp.size() == 4);
	assert(exp[0] == 'S' || exp[0] == 'U');

	// ccdNr() below returns the CCD number for the X-ray instruments,
	// but the window number for OM. In the comments I always use CCD,
	// but in the messages and in the summary file I use `word'.
	string word = "CCD";
	if(inst == "OM") 
	{
		word = "window";
	}
  
	std::map<unsigned, Xmm::DataMode> dmodes; // map CCD number to CCD data mode
	std::map<unsigned, Xmm::Mode> cmodes; // map CCD number to CCD calibration mode

	std::set<Xmm::Mode> ss; // track how many unique calibration modes we have seen
	Xmm::Mode instmode = Xmm::MODE_NONE;

	// This block of code was moved down from below the main block of code in order to 
	// use the Ippvs in mode determination
	vector<STime> v = exposure_start_end_times(inst,exp);

	STime start = v[0];
	STime stop = v[1];

	IppvExtractorServer mes = IppvExtractorServer();
	IppvExtractor *me = mes.extractor(Xmm::toInstrument(fits_names[inst]), _dir, expid(inst,exp));
	
	STime t1(start);
	STime t2(stop);

	if(stop - start > 2*_tpad)
	{
		t1 += _tpad;
		t2 -= _tpad;
	} 

	map<string,string> ippvValues = me->values(t1, t2);

	// The determination of the instrument modes requires access to the
	// CAL, hence the CCF. There is a minimal CCF that will make the CAL
	// happy for this, namely one with only the MODEPARAMS files. This
	// is however not guaranteed to be for ever true. It is mentioned
	// here in case during testing the full CCF is not available.	

	bool disable_mode_determination = false;
	
	// SCR117 Flag if any of the data files corresponds to a science mode,
	// in which case the standard treatment applied
	bool isScience = false;
	
	if(_find_instrument_modes)
	{
		Directory d(_dir);
		d.filter("????_??????????_" + inst + exp + "?????.FIT");

		// If we are here is because we know that there must be some
		// dataset. So it is ok for this to remain an error.
		if(d.size() == 0) {
			errstr << "Looking for one or more files matching the pattern "
				<< d.filter() << " in the directory "
				<< _dir << ". I found " << d.size() << " instead. From a previous scan I expected to find at least one file."
				<< Sas::error("IncorrectNumberOfDatasets");
		}
		
		errstr << "Determination of the data modes for "
			<< expid(inst,exp)
			<< ". Found " << d.size() << " ODF constituents."
			<< Sas::message(Msg::App, Msg::Noisy);
		cmodes.clear();
		for(Directory::const_iterator k = d.begin(); k != d.end(); ++k)
		{

			OdfSet s(_dir + "/" + (*k)->basename());
			errstr << s.basename() << " has data mode " << s.dataMode() <<
				Sas::message(Msg::App, Msg::Noisy);

			// FIXME: this whole section needs to be reviewed as it is
			// becoming a complete mess.

			//        // For OM, skip all the auxiliary datasets, as they refer to the
			//        // whole of the exposure and they throw off balance the
			//        // instrument mode determination algorithm.
			//        if(s.dataMode() == Xmm::PRIORITY_WINDOW_DATA ||
			//  	 s.dataMode() == Xmm::REFERENCE_FRAME ||
			//  	 s.dataMode() == Xmm::PRIORITY_FIELD_ACQUISITION ||
			//  	 s.dataMode() == Xmm::PRIORITY_FAST ||
			//  	 s.dataMode() == Xmm::TRACKING_HISTORY)
			//  	continue;


			// We ask about the CCD data mode only if we're dealing with a
			// component that is science related. However, not all
			// science-related ODF components lead to a camera mode (for
			// instance RGS diagnostic images) and that we catch by asking
			// the calServer if it could determine the mode.
			if(s.type() == OdfSet::EVENT_LIST || s.type() == OdfSet::IMAGE)
			{

				// Do I have already a data mode entry for the current CCD? If
				// I do then I expect it to be the same as the one I already have.
				const Xmm::DataMode current_dmode = s.dataMode();

				// The following condition could only happen for EMOS, CCD 1
				// for which a two-node readout mode (leading to two ODF
				// constituents for one CCD) is possible. In this case I
				// expect the CCD data mode to be the same. 

				// WARNING: in at least one case I have seen:
				//
				// 0115_0127720201_PNU00201IME.FIT
				// 0115_0127720201_PNU00201ODI.FIT
				//
				// This triggers the warning.
				//
				// 2001-10-30: Nope, this condition is normal, in that for
				// some configurations two types of data are being collected
				// at the same time, hence two types of files will appear with
				// the same exposure number.
				//
				// 2000-10-10: there seems to be agreement that such a
				// condition should not happen (ie, it is a bug in the
				// ODS). It seems that its occurrence is related to some
				// manual telecommanding (exposures are unscheduled) that
				// fools the software that should put the data into two
				// different exposures.
				//
				// FIXME: hack to work around the PN ODI + IME data
				// types. Note that here I rely on how the dataset names sort,
				// so IMI ahead of ODI. When the condition fails I only test
				// for ODI.
				//
				// FIXME: same hack for NOI

				// FIXME: 2001-11-06 auch, I can also have TIE + ODI/NOI

				// FIXME: 2002-01-06 we can expect BUE + NOI/ODI as well


				// FIXME: 2020-01-17 and I can have DLI + IME as well.

				// FIXME: 2004-07-07 Add HDI + IME

				if(dmodes.find(s.ccdNr()) != dmodes.end())
				{
					// I already have an entry for the current CCD
					if(current_dmode == Xmm::OFFSET_DATA || 
						current_dmode == Xmm::HBR_OFFSET_DATA ||
						current_dmode == Xmm::NOISE_DATA ||
						current_dmode == Xmm::DISCARDED_LINES_DATA)
					{
						if ( dmodes[s.ccdNr()] != Xmm::DISCARDED_LINES_DATA )
						{
							// The new entry is ODI or NOI => ignore it
							errstr << "Looking at the data mode of " << s.basename()
								<< " I found already an entry for " + word
								<< s.ccdNr() << " The entry I have is "
								<< dmodes[s.ccdNr()] << " whereas the current one is "
								<< current_dmode << " The latter will be ignored."
								<< Sas::message(Msg::AppLib, Msg::Verbose);
							continue;
						}

					}

					if(dmodes[s.ccdNr()] == Xmm::OFFSET_DATA || 
						dmodes[s.ccdNr()] == Xmm::HBR_OFFSET_DATA ||
						dmodes[s.ccdNr()] == Xmm::NOISE_DATA ||
						dmodes[s.ccdNr()] == Xmm::DISCARDED_LINES_DATA)
					{
						// The entry I already have is ODI or NOI or DLI => replace
						// it with the current one.
						errstr << "Looking at the data mode of " << s.basename()
							<< " I found already an entry for " + word
							<< s.ccdNr() << " The entry I have is "
							<< dmodes[s.ccdNr()] << " whereas the current one is "
							<< current_dmode << " The former will be replaced."
							<< Sas::message(Msg::AppLib, Msg::Verbose);
						dmodes[s.ccdNr()] = current_dmode;

					}

				}
				
				if(dmodes.find(s.ccdNr()) != dmodes.end() &&
					dmodes[s.ccdNr()] != current_dmode)
				{
					errstr << "Found inconsistent duplicate " + word + " data mode entry for "
						<< s.basename() << " " + word + " number " << s.ccdNr()
						<< ". The existing entry is: "
						<< dmodes[s.ccdNr()] << " the one I found is "
						<< current_dmode << endl
						<< "All operations related to the determination of the CCD data mode, calibration mode, and instrument configuration will be disabled for the current exposure."
						<< Sas::warning("InconsistentDataMode");
					disable_mode_determination = true;
					break; // no need to continue loping over the ODF constituents
				} 
				else
				{
					dmodes[s.ccdNr()] = current_dmode;
				}

				// FIXME
				// 2001-10-30: here I should use
				// std::vector<Xmm::DataModes>& science_data_modes = 
				// 	xmmPayLoad.instrument(s.instrument())
				// For these modes the cal is able to return an answer. This
				// I'll be able to do when the releasetrack version of
				// odfingest meets the development version of cal/caloalutils.

				// Only CCD data modes that are science data modes (as opposed
				// to science related) are processed to determine the camera
				// mode. 
				vector<Xmm::DataMode>::const_iterator p =
					find(science_data_modes.begin(), science_data_modes.end(),
						current_dmode);

				// SCR117 - let the CAL attempt to work out what mode it is
				// regardless of whether it is science or not
				if(p == science_data_modes.end())
				{
				
					errstr << s.dataMode() << " is not a science data mode."
						<< Sas::message(Msg::App, Msg::Noisy);	  
				}
				else {
					isScience = true;					
				}

				Xmm::Mode calibration_mode = Xmm::MODE_NONE;
				// set the CAL state, but ask the server to be silent about
				// a failure to determine the mode and the filter.

				calServer.setState(s.name(), true, true);
				CalState *cs = calServer.state();
				ChoiceStateVariable<Mode> *m = cs->mode();

				if(m->isUncertain()) 
				{
					errstr << "I asked the calibration access layer to determine the data mode for "
						<< s.basename() << " and I did not get a unique answer."
						<< "All operations related to the determination of the CCD data mode, calibration mode, and instrument configuration will be disabled for the current exposure."
						<< Sas::warning("UnknownCalibrationMode");
					disable_mode_determination = true;
					break; // no need to continue looping over the ODF constituents	
				} 
				else 
				{ 
					calibration_mode = m->get();

					// SCR 117, need the same checks as above for data modes
					// since the CAL will return non-science modes
					if(cmodes.find(s.ccdNr()) != cmodes.end())
					{
						// I already have an entry for the current CCD
						if(calibration_mode == Xmm::OFFSET || 
							calibration_mode == Xmm::NOISE ||
							calibration_mode == Xmm::DIAGNOSTIC)
						{
							if (cmodes[s.ccdNr()] != Xmm::DIAGNOSTIC)
							{
								// The new entry is ODI or NOI => ignore it
								errstr << "Looking at the calibration mode of " << s.basename()
									<< " I found already an entry for " + word
									<< s.ccdNr() << " The entry I have is "
									<< cmodes[s.ccdNr()] << " whereas the current one is "
									<< calibration_mode << " The latter will be ignored."
									<< Sas::message(Msg::AppLib, Msg::Verbose);
								continue;
							}

						}

						if(cmodes[s.ccdNr()] == Xmm::OFFSET || 
							cmodes[s.ccdNr()] == Xmm::NOISE ||
							cmodes[s.ccdNr()] == Xmm::DIAGNOSTIC)
						{
							errstr << "Looking at the calibration mode of " << s.basename()
								<< " I found already an entry for " + word
								<< s.ccdNr() << " The entry I have is "
								<< cmodes[s.ccdNr()] << " whereas the current one is "
								<< calibration_mode << " The former will be replaced."
								<< Sas::message(Msg::AppLib, Msg::Verbose);

							// Remove the mode from the set of identified modes
							ss.erase(cmodes[s.ccdNr()]);
							ss.insert(calibration_mode);

							// Update the mode for that CCD

							cmodes[s.ccdNr()] = calibration_mode;

						}

					}

					if(cmodes.find(s.ccdNr()) != cmodes.end() &&
						cmodes[s.ccdNr()] != calibration_mode)
					{

						// This could only happen for EMOS, CCD 1 for which a two-node
						// readout mode is possible. Duplicate entries are ok, but the I
						// expect the CCD camera mode to be the same.
						//
						// But see WARNING above.
						errstr << "Found inconsistent duplicate " + word + " camera mode entry for "
							<< s.basename() << " " + word + " number " << s.ccdNr()
							<< ". The existing entry is: "
							<< cmodes[s.ccdNr()]
							<< "All operations related to the determination of the CCD data mode, calibration mode, and instrument configuration will be disabled for the current exposure."
							<< Sas::warning("InconsistentCalibrationMode");
						disable_mode_determination = true;
						break; // no need to continue looping over the ODF constituents
					} 
					else 
					{

						cmodes[s.ccdNr()] = calibration_mode;
						ss.insert(calibration_mode);
						errstr << s.basename() << " has calibration mode " << calibration_mode
							<< Sas::message(Msg::App, Msg::Verbose);
					} // good calibration mode
				} // uncertain calibration mode
			} // is an event list or an image
		} // for each ODF constituent 
  
		if(disable_mode_determination)
		{
			cmodes.clear();
			dmodes.clear();
		}



		// Instrument mode determination:
		// A fair amount of mode-remapping happens here, so it get messy
		// A longer term solution is to switch to nodfingest, which may
		// be more maintainable
		
		// Non-Science Modes
		// OM: Special Case: Check for Engineering 4 with FW closed, could be Dark or Flat
		//		
		// Science Modes:
		// MOS: the CAL camera mode for CCD 1 (either node). Expect to
		// see at most 2 data modes (one for CCD 1, one for the other 6).
		//
		// PN: the CAL camera mode for any CCD (they all operate
		// in the same manner). Expect to see only 1 data mode.
		// PN Special Case: Check Frame Time Parameter to see if mode
		// was Extended Full Frame
		// PN Special Case: Check CAMEX for CCD4 to see if mode
		// was Low Gain
		// PN Special Case: Check MASK_MODE for CCD4 to see if mode
		// was Masked (RGS or PN Prime)
		//
		// RGS: the CAL camera mode for any CCD (they all operate
		// in the same manner). 
		// RGS Special Case: If the mode for all CCDs is HTR-SINGLE
		// the calibration mode is HTR-MULTIPLE.
		// RGS Special Case: If Mode == Spectroscopy, but OCB == 1 and CSG == 10,
		// the calibration mode is Readout Storage Section
		// RGS Special Case: If Mode == HER+SES, but OCB == 1 and CSG != 10,
		// the calibration mode is Spectroscopy 1x1
		//
		// OM: the CAL camera mode for window 0. Expect to see at most 2
		// data modes (IMI with FAE).
		// OM Special Case:	Engineering 2 with FW closed, could be Dark or Flat
		
		if(cmodes.size() != 0)
		{

			// SCR117 If no science files, just use the first calibration mode
			// at the moment; lets see what wierd cases emerge
			if (!isScience) {

				string ledLevel = ippvValues["FLOOD_LED_BIAS"];

				switch(*ss.begin()) {
					case  Xmm::DARK_HIGH:

						if (ledLevel != "SAS_UNDEFINED") 
						{
							if (ledLevel != "0") 
							{
								instmode = Xmm::FLAT_FIELD_HIGH;	
							}
							else 
							{						
								instmode = Xmm::DARK_HIGH;												
							}
						}
						else {
							instmode = Xmm::DARK_HIGH;						
						}
						break;				
				default:
						instmode = *ss.begin();							
				}
			}
			else 
			{

				if(inst == "M1" || inst == "M2")
				{ // MOS 
					if(ss.size() > 2)
					{
						errstr << "While looking at " << expid(inst,exp)
							<< " I found " << ss.size()
							<< " data mode entries, whereas I was expecting no more than 2. Thereby I cannot determine the instrument mode."
							<< Sas::warning("UnknownInstrumentMode");      
						instmode = Xmm::MODE_NONE;
					} 
					else
					{
						// for MOS the lowest CCD number is 1
						if(cmodes.find(1) != cmodes.end()) 
						{
							instmode = cmodes[1];
						}
						else 
						{
							errstr << "While looking at " << expid(inst,exp)
								<< " I asked " + word + " " 
								<< " 1 to determine the instrument mode. Apparently there is no corresponding data. Thereby I cannot determine the instrument mode." << Sas::warning("UnknownInstrumentMode"); 
							instmode = Xmm::MODE_NONE;
						}
					}
				} 
				else if(inst == "OM")
				{ // OM 
					if(ss.size() > 2)
					{
						errstr << "While looking at " << expid(inst,exp)
							<< " I found " << ss.size()
							<< " data mode entries, whereas I was expecting no more than 2. Thereby I cannot determine the instrument mode."
							<< Sas::warning("UnknownInstrumentMode");      
						instmode = Xmm::MODE_NONE;
					} 
					else
					{

						string ledLevel = ippvValues["FLOOD_LED_BIAS"];

						// For OM the lowest window number is 0, 
						if(cmodes.find(0) != cmodes.end()) 
						{
							switch (cmodes[0]) 
							{
								case  Xmm::DARK_LOW:

									if (ledLevel != "SAS_UNDEFINED") 
									{
										if(ledLevel != "0") 
										{
											instmode = Xmm::FLAT_FIELD_LOW;												
										}
										else {						
											instmode = Xmm::DARK_LOW;							
										}
									}
									else 
									{
										instmode = Xmm::DARK_LOW;													
									}
									break;
								default:
									instmode = cmodes[0];
							}
						}
						else 
						{
							errstr << "While looking at " << expid(inst,exp)
								<< " I asked " + word + " " 
								<< " 0 to determine the instrument mode. Apparently there is no corresponding data. Thereby I cannot determine the instrument mode." << Sas::warning("UnknownInstrumentMode"); 
							instmode = Xmm::MODE_NONE;
						}
					}
				} 
				else if (inst == "R1" || inst == "R2") 
				{ // RGS
					// If the mode is HTR-MULTI
					if (*ss.begin() == Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD) {
						// But only 1 CCD
						if(cmodes.size() == 1) {

							// Update the calibration mode

							// Get the CCD number
							std::map<unsigned, Xmm::Mode>::const_iterator thismode = cmodes.begin();
							unsigned ccd = (*thismode).first;
							
							// Remove existing configuration
							cmodes.clear();					
							
							// Set the configuration mode
							cmodes[ccd] = Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD;	
							
							// Set the instrument mode to HTR-SINGLE
							instmode = Xmm::HIGH_TIME_RESOLUTION_SINGLE_CCD;
						}
						else {
							// Set the instrument mode to HTR-MULTIPLE
							instmode = Xmm::HIGH_TIME_RESOLUTION_MULTIPLE_CCD;
						}
					}
					else if (*ss.begin() == Xmm::SPECTROSCOPY) {
							
						string OCB = ippvValues["CCD_9_OCB"];
						string CSG = ippvValues["CCD_9_CSG_PATTERN"];

						if ((OCB != "") && (CSG != "") &&
							(OCB != "SAS_UNDEFINED") && (CSG != "SAS_UNDEFINED") &&
							(OCB == "1") && (CSG == "10")) {

							instmode = Xmm::READOUT_STORAGE_SECTION;

						}
						else {
						
							instmode = Xmm::SPECTROSCOPY;
							
						}
					}
					else if (*ss.begin() == Xmm::HIGH_EVENT_RATE_WITH_SES) {

						string OCB = ippvValues["CCD_9_OCB"];
						string CSG = ippvValues["CCD_9_CSG_PATTERN"];

						if ((OCB != "") && (CSG != "") &&
							(OCB != "SAS_UNDEFINED") && (CSG != "SAS_UNDEFINED") &&
							(OCB == "1") && (CSG != "10")) {

							instmode = Xmm::READOUT_STORAGE_SECTION;

						}
						else {
						
							instmode = Xmm::HIGH_EVENT_RATE_WITH_SES;
							
						}
					}
					else if (ss.size() != 1) {
						errstr << "While looking at " << expid(inst,exp)
							<< "I found " << ss.size()
							<< " data mode entries, whereas I was expecting only 1."
							<< Sas::warning("UnknownInstrumentMode");
						instmode = Xmm::MODE_NONE;
					} 
					else {
						instmode = *ss.begin();
					}
				}
				else
				{ // PN
					if (*ss.begin() == Xmm::PRIME_FULL_WINDOW) {
					
						string ftp = ippvValues["FRAME_TIME_PARAMETER"];
						string maskP = ippvValues["A1_CMLINE_PIXSET"];
						string maskC = ippvValues["A1_CMLINE_CCD"];
						string maskN = ippvValues["A1_CMLINE_NUMB"];
						string camex4 = ippvValues["CAMEX_GAIN_CCD4"];

						instmode = Xmm::PRIME_FULL_WINDOW;

						if ((ftp != "SAS_UNDEFINED") && (ftp != "") && (ftp != "0")) {
							instmode = Xmm::PRIME_FULL_WINDOW_EXTENDED;
						}
						
						if ((maskP != "") && (maskC != "") && (maskN != "") &&
							(maskP != "SAS_UNDEFINED") && (maskC != "SAS_UNDEFINED") && (maskN != "SAS_UNDEFINED") &&
							(maskP == "BAD") && (maskC == "1") && (maskN == "1")) {
						
							instmode = Xmm::PRIME_FULL_WINDOW_MASKED;
						} 
						
						if ((camex4 != "SAS_UNDEFINED") && (camex4 != "") && (camex4 != "HIGH")){
							instmode = Xmm::PRIME_FULL_WINDOW_LOWGAIN;
						} 

					}
					else {
		    			if(ss.size() != 1)
						{
							errstr << "While looking at " << expid(inst,exp)
								<< "I found " << ss.size()
								<< " data mode entries, whereas I was expecting only 1."
								<< Sas::warning("UnknownInstrumentMode");
							instmode = Xmm::MODE_NONE;
						} 
						else 
						{
							instmode = *ss.begin();
						}
					}
				} // instrument mode determination: MOS vs non MOS
			} // is scientific mode?
		} // instrument mode determination
	} // was asked to determine the instrument mode

	IppvExtractorServer ies = IppvExtractorServer();
	IppvExtractor *ie = ies.extractor(Xmm::toInstrument(fits_names[inst]), _dir, expid(inst,exp));

	os << "//" << endl;
	os << "// Configuration Record for " << inst << endl;
	os << "//" << endl;
	os << "CONFIGURATION " + long_names[inst] << endl;

	unsigned number_of_parameters = 1; 	// will always write the
										// instrument mode, even if set to Xmm::MODE_NONE
										
	const bool write_modes = _find_instrument_modes &&
		!disable_mode_determination;

	if(write_modes) 
	{
		number_of_parameters += cmodes.size() + dmodes.size();
	}

	bool extract_ippv = _use_housekeeping;
	
	for(map<unsigned,Xmm::DataMode>::const_iterator k = dmodes.begin();
		k != dmodes.end(); ++k)
	{
		if(extract_ippv && find(_no_ippv_data_modes.begin(),
			    _no_ippv_data_modes.end(),
			    (*k).second) != _no_ippv_data_modes.end())
		{
			extract_ippv = false;
			errstr << "For the data mode " << (*k).second
				<< " no housekeeping data will be extracted." <<
				Sas::message(Msg::App, Msg::Verbose);
		}
	}


	// JSH - We are going to omit HBR_OFFSET_DATA, if this exposure ID/sched flag is identical
	// to another non-science exposre
	// So let count the number of records we have. 

	// Why are we going to omit them, well, they are introduced
	// retrospectively into the ODF through QCheck, and there are a number of
	// ODFs (for example 0164560401) where the HDI and DLI exposures IDs clash, causing
	// confusion. Since HDI files are attached to science data, they should not exist
	// independently of the science data and therefore should not appear as individual
	// DATA_MODE entries in the summary file. The OAL provides access to the files
	// independently of the Summary File (for the above reason).

	int n_hbr_dumps = 0;    
	bool exclude_hbr = false;

	for(map<unsigned,Xmm::DataMode>::const_iterator k = dmodes.begin();
		k != dmodes.end(); ++k) 
	{
    
		if ((*k).second == Xmm::HBR_OFFSET_DATA) 
		{
			n_hbr_dumps++;
		}
    
		if((*k).second == Xmm::OFFSET_DATA || 
			(*k).second == Xmm::NOISE_DATA ||
			(*k).second == Xmm::DISCARDED_LINES_DATA) 
		{    
			exclude_hbr = true;      
		}    
	}
  
	if(extract_ippv) 
	{
		number_of_parameters += ie->ippvs();
	}

	if(exclude_hbr) 
	{
		number_of_parameters -= n_hbr_dumps;
	}

	os << number_of_parameters << " / Number of configuration parameters available"
		<< endl;
		
	os << "MODE = " << Xmm::toString(instmode) << " / Instrument configuration "
		<< endl;
		
	if(write_modes)
	{

		for(map<unsigned,Xmm::Mode>::const_iterator k = cmodes.begin();
		k != cmodes.end(); ++k) 
		{
			os << "CALIBRATION_MODE_" << (*k).first << " = " << Xmm::toString((*k).second)
				<< " / Mode used to calibrate events from " + word + " " << (*k).first
				<< endl;
		}

		for(map<unsigned,Xmm::DataMode>::const_iterator k = dmodes.begin();
		k != dmodes.end(); ++k)
		{
			if ( ((*k).second != Xmm::HBR_OFFSET_DATA) ||
			   (((*k).second == Xmm::HBR_OFFSET_DATA) && !exclude_hbr)) 
			{

				os << "DATA_MODE_" << (*k).first << " = " << Xmm::toString((*k).second)
					<< " / Data mode for " + word + " " << (*k).first
					<< endl;
			}
		}
	}

	if(extract_ippv)
	{
		// Periodic HK packets are sent to the ground every 8s. It is
		// possible that a parameter is changed, the observation started,
		// but the new value is not yet available in the telemetry. 

		// See note about _tpad at the constructor level.

		STime t1(start);
		STime t2(stop);

		if(stop - start > 2*_tpad)
		{
			t1 += _tpad;
			t2 -= _tpad;
		} 
		else 
		{
			errstr << expid(inst,exp) << " is shorter than " << 2*_tpad
				<< "s. The extraction of housekeeping parameters may fail. (In which case warnings will be issued.)"
				<< Sas::message(Msg::App, Msg::Verbose);    
		 }

		ie->write(os, t1, t2);
	
	}  
	ENDPROC
}

void OdfObject::proposal_summary_rec(ostream & os) const {
	PROC
	errstr << "Writing the proposal summary record." <<
		Sas::message(Msg::App, Msg::Noisy);

	// FIXME: code duplication. Also in configuration_details_rec()
	Directory d(_dir);
	d.filter(icdosf());

	if(d.size() != 1) 
	{
		errstr << "Looking for one and only one file named " << icdosf()
			<< " in the directory " << _dir << ". Instead I found "
			<< d.size() << Sas::error("IncorrectNumberOfDatasets");
	}
	
	ifstream is((*d.begin())->name().c_str());
	string line;
	// Goes through the file until it reaches the PROPOSAL section
	while(getline(is,line) && line.substr(0,8) != "PROPOSAL" && is.good());

	if(!is.eof())
	{
		os << "// The PROPOSAL record was extracted from " << (*d.begin())->basename()
			<< endl;
		os << line << endl;

		// Now copy all lines until the end of the file
		while(!is.eof())
		{
			getline(is,line);
			os << line << endl;;
		}
	} 
	else
	{
		Sas::warning("MissingData", "No PROPOSAL record found. This does not stop you from analyzing your data. If it causes an error at a later stage in the analysis, consider adding your own fake proposal record to the SUM.SAS file.");
		is.close();
	}

	ENDPROC
}

const bool& OdfObject::valid() const { PROC RETURN _valid; }


